// FtpSend.cpp : Implementation file for MQSI V2.0.1 plug-in node
//
// Template built by MQSI V2 Visual Studio Wizard (kolban@us.ibm.com)
//
// Installation:
// For NT:		<MQSI Root>\bin
// For Unix:	<MQSI Root>/lil
//

#include "stdafx.h"

#include <BipCci.h>		// Common functions (between nodes and parsers)
#include <BipCni.h>		// Message processing nodes
#include <stdarg.h>		// Required for variable length arguments
#include <stdio.h>		// Required for file IO and NULL declaration
#include <stdlib.h>		// Required for memory manipulation

#include "NodeUtils.h"	// Include the required utility functions
#include "msg/FtpSend_msg.h"

#ifdef _WIN32
HMODULE	g_module;
BOOL APIENTRY DllMain(
	HMODULE hModule,
	DWORD  ul_reason_for_call,
	LPVOID lpReserved
					 )
{
	g_module = hModule;
    return TRUE;
}

void __declspec(dllexport) *bipGetMessageflowNodeFactory();

#endif


///////////////////////////////////////////////////////////
// NodeContext
struct NodeContext
{
	MQSIX_ATTRIBUTES	*attributes;
	CciTerminal	*inTerminal;						// In Terminal
	CciTerminal	*outTerminal;						// Out Terminal
	CciTerminal	*failureTerminal;					// Out Terminal
	char		strFtpServerName[MQSIX_MAX_STRING];
	char		strFtpUserid[MQSIX_MAX_STRING];
	char		strFtpPassword[MQSIX_MAX_STRING];
	char		strFtpFileName[MQSIX_MAX_STRING];
	int			isAscii;
};



///////////////////////////////////////////////////////////
// CreateNodeContext
void *CreateNodeContext(
	CciFactory	*factoryObject,
	CciChar		*nodeName,
	CciNode		*nodeObject)
{
	struct NodeContext	*pMyContext;
	int					iRc;

	pMyContext = (struct NodeContext *)malloc(sizeof(struct NodeContext));

	mqsixLog(">> CreateNodeContext: Node Name=%S\n", nodeName);
	pMyContext->inTerminal = cniCreateInputTerminal(&iRc, nodeObject, L"in");
	pMyContext->outTerminal = cniCreateOutputTerminal(&iRc, nodeObject, L"out");
	pMyContext->failureTerminal = cniCreateOutputTerminal(&iRc, nodeObject, L"failure");
	pMyContext->attributes = NULL;
	pMyContext->isAscii = 0;
	mqsixAddAttribute(&pMyContext->attributes, L"ftpServerName", L"none");
	mqsixAddAttribute(&pMyContext->attributes, L"ftpUserid", L"none");
	mqsixAddAttribute(&pMyContext->attributes, L"ftpPassword", L"none");
	mqsixAddAttribute(&pMyContext->attributes, L"ftpFileName", L"none");
	mqsixAddAttribute(&pMyContext->attributes, L"ftpAscii",	L"false");

	return(pMyContext);
}

///////////////////////////////////////////////////////////
// DeleteNodeContext
void DeleteNodeContext(CciContext *context)
{
	struct NodeContext	*pMyContext;

	pMyContext = (struct NodeContext *)context;
	mqsixLog(">> DeleteNodeContext\n");
	mqsixReleaseAttributes(pMyContext->attributes);
	pMyContext->attributes = NULL;
	free(context);
}

///////////////////////////////////////////////////////////
// GetAttributeName
int GetAttributeName(
	CciContext	*context,
	int			index,
	CciChar		*buffer,
	int			bufsize)
{
	int	iRc;
	struct NodeContext	*pMyContext;

	pMyContext = (struct NodeContext *)context;
	iRc = mqsixGetAttributeName(
		pMyContext->attributes,
		index,
		buffer,
		bufsize);
	return(iRc);
}

///////////////////////////////////////////////////////////
// GetAttribute
int GetAttribute(
	CciContext	*context,
	CciChar		*attrName,
	CciChar		*buffer,
	int bufsize)
{
	struct NodeContext	*pMyContext;
	int					iRc;

	pMyContext = (struct NodeContext *)context;
	iRc = mqsixGetAttribute(
		pMyContext->attributes,
		attrName,
		buffer,
		bufsize);
	return(iRc);
}

///////////////////////////////////////////////////////////
// SetAttribute
int SetAttribute(
	CciContext	*context,
	CciChar		*attrName,
	CciChar		*attrValue)
{
	struct NodeContext	*pMyContext;
	int					iRc;
	char				*strTemp;

	pMyContext = (struct NodeContext *)context;
	iRc = mqsixSetAttribute(pMyContext->attributes,
		attrName,
		attrValue);

	if (mqsixStringCompare(attrName, L"ftpServerName") == 0)
	{
		strTemp = mqsixUnicodeToASCII(attrValue);
		strcpy(pMyContext->strFtpServerName, strTemp);
		free(strTemp);
	}
	if (mqsixStringCompare(attrName, L"ftpUserid") == 0)
	{
		strTemp = mqsixUnicodeToASCII(attrValue);
		strcpy(pMyContext->strFtpUserid, strTemp);
		free(strTemp);
	}
	if (mqsixStringCompare(attrName, L"ftpPassword") == 0)
	{
		strTemp = mqsixUnicodeToASCII(attrValue);
		strcpy(pMyContext->strFtpPassword, strTemp);
		free(strTemp);
	}
	if (mqsixStringCompare(attrName, L"ftpFileName") == 0)
	{
		strTemp = mqsixUnicodeToASCII(attrValue);
		strcpy(pMyContext->strFtpFileName, strTemp);
		free(strTemp);
	}
	// Check for the Ascii/Binary flag and set boolean appropriately
	if (mqsixStringCompare(attrName, L"ftpAscii") == 0)
	{
		if (mqsixStringCompare(attrValue, L"true") == 0)
		{
			pMyContext->isAscii = 1;
		}
		else
		{
			pMyContext->isAscii = 0;
		}
	}
	return(iRc);
}

///////////////////////////////////////////////////////////
// FtpSend
int FtpSend(struct NodeContext *pMyContext, char *pFileNameA, const CciByte *pData, CciSize iLength)
{
	HINTERNET	hInternet;
	HINTERNET	hFtp;
	HINTERNET	hFile;
	DWORD		iError;
	DWORD		iContext = 123;
	DWORD		iWritten;
	int			iRc;
	char		strError[20];
	char		strErrorText[512];
	DWORD		iErrorInternet;
	DWORD		iTextLength;

	mqsixLog("Ftping to: server=%s, userid=%s, password=%s, file=%s, length=%d\n",
		pMyContext->strFtpServerName,
		pMyContext->strFtpUserid,
		pMyContext->strFtpPassword,
		pFileNameA,
		iLength);

	hInternet = InternetOpen("MQSI V2.0.1", 
		INTERNET_OPEN_TYPE_DIRECT,
		NULL,
		NULL,
		0);
	if (hInternet == NULL)
	{
		iError = GetLastError();
		mqsixLog("Error with InternetOpen!: %d\n", iError);
		return(1);
	}

	hFtp = InternetConnect(hInternet, pMyContext->strFtpServerName,
		INTERNET_DEFAULT_FTP_PORT,
		pMyContext->strFtpUserid, pMyContext->strFtpPassword,
		INTERNET_SERVICE_FTP, 0, iContext);
	if (hFtp == NULL)
	{
		iError = GetLastError();
		switch(iError)
		{
		case ERROR_INTERNET_NAME_NOT_RESOLVED:
			cciLog(&iRc, CCI_LOG_ERROR, __FILE__, __LINE__, "FtpSend",
				L"MQSIV2_FTPSEND", MSG_INTERNET_NAME_NOT_RESOLVED,
				"(%1) Unable to resolve hostname %2",
			pMyContext->strFtpServerName, (char *)0);
			break;
		case ERROR_INTERNET_TIMEOUT:
			cciLog(&iRc, CCI_LOG_ERROR, __FILE__, __LINE__, "FtpSend",
				L"MQSIV2_FTPSEND", MSG_INTERNET_ERROR_INTERNET_TIMEOUT,
				"(%1) Timed out trying to connect to FTP server at '%2'",
			pMyContext->strFtpServerName, (char *)0);
			break;
		case ERROR_INTERNET_INCORRECT_PASSWORD:
		case ERROR_INTERNET_INCORRECT_USER_NAME:
			cciLog(&iRc, CCI_LOG_ERROR, __FILE__, __LINE__, "FtpSend",
				L"MQSIV2_FTPSEND", MSG_INTERNET_ERROR_BAD_USERID_PASSWORD,
				"(%1) Invalid userid/password trying to connect to FTP server '%2'",
			pMyContext->strFtpServerName, (char *)0);
			break;
		default:
			sprintf(strError, "%d", iError);
			cciLog(&iRc, CCI_LOG_ERROR, __FILE__, __LINE__, "FtpSend",
				L"MQSIV2_FTPSEND", MSG_INTERNET_ERROR_OTHER,
				"(%1) Internet Connect error %2 trying to connect to server at '%3'",
			strError, pMyContext->strFtpServerName, (char *)0);
			break;
		}
		mqsixLog("Error with InternetConnect!: %d\n", iError);
		InternetCloseHandle(hInternet);
		return(1);
	}

	hFile = FtpOpenFile(hFtp, 
		pFileNameA,
		GENERIC_WRITE,
		pMyContext->isAscii?FTP_TRANSFER_TYPE_ASCII:FTP_TRANSFER_TYPE_BINARY,
		0);
	if (hFile == NULL)
	{
		iError = GetLastError();
		if (iError == ERROR_INTERNET_EXTENDED_ERROR)
		{
			iTextLength = sizeof(strErrorText);
			if (InternetGetLastResponseInfo(&iErrorInternet, strErrorText, &iTextLength))
			{
				cciLog(&iRc, CCI_LOG_ERROR, __FILE__, __LINE__, "FtpSend",
					L"MQSIV2_FTPSEND", MSG_INTERNET_ERROR_EXTENDED,
					"(%1) Error response from FTP server '%2'\n%3",
					pMyContext->strFtpServerName, strErrorText,(char *)0);
			}
		}
		mqsixLog("Error with FtpOpenFile!: %d\n", iError);
		InternetCloseHandle(hInternet);
		InternetCloseHandle(hFtp);
		return(1);
	}
	if (!InternetWriteFile(hFile, pData, iLength, &iWritten))
	{
		iError = GetLastError();
		mqsixLog("Error with InternetWriteFile!: %d\n", iError);
	}
	InternetCloseHandle(hFile);
	InternetCloseHandle(hFtp);
	InternetCloseHandle(hInternet);
	return(0);

}

///////////////////////////////////////////////////////////
// Evaluate
void Evaluate(
	CciContext	*context, 
	CciMessage	*inDestinationList,
	CciMessage	*inExceptionList,
	CciMessage	*inMessage)
{
	int					iRc;
	struct NodeContext	*pMyContext;
	const CciByte				*pData;
	CciSize				iSize;
	CciMessage			*pTempMessage;
	CciElement			*pRoot;
	CciElement			*pMQMDElement;
	CciElement			*pFileNameElement;

	CciChar				*pFileNameU;
	char				*pFileNameA;
	int					iFtpRc;

	mqsixLog(">> Evaluate\n");
	pMyContext = (struct NodeContext *)context;

	mqsixLog("FTPing the output data ...\n");
	pTempMessage = cniCreateMessage(&iRc, cniGetMessageContext(&iRc, inMessage));
	pRoot = cniRootElement(&iRc, inMessage);
	cniCopyElementTree(&iRc, pRoot, cniRootElement(&iRc, pTempMessage));
	// Delete the MQMD subtree ...
	pMQMDElement = cniSearchFirstChild(&iRc, cniRootElement(&iRc, pTempMessage),
		CCI_COMPARE_MODE_NAME, 0, L"MQMD");
	if (pMQMDElement != NULL)
	{
		cniDetach(&iRc, pMQMDElement);
	}
	cniWriteBuffer(&iRc, pTempMessage);
	if (iRc != 0)
	{
		mqsixLog("Error from cniWriteBuffer(): %d\n", iRc);
	}
	else	// No error from cniWriteBuffer()
	{

	// Support the capability of the message flow supplying the file name.  This
	// will over-ride any hard-coded name that is supplied in the node's attributes
	// The file name will be taken from the "Root.WKSPACE.FTPSend.FileName" location
	// in the incoming message tree.
		pFileNameA = NULL;
		pFileNameElement = mqsixFindElement("Root.WKSPACE.FTPSend.FileName", pRoot);
		if (pFileNameElement != NULL)
		{
			mqsixElementCharValue(pFileNameElement, &pFileNameU);
			if (pFileNameU != NULL)
			{
				pFileNameA = mqsixUnicodeToASCII(pFileNameU);
				free(pFileNameU);
			}
		}
		mqsixLogTree(cniRootElement(&iRc, pTempMessage));
		iSize = cniBufferSize(&iRc, pTempMessage);
		pData = cniBufferPointer(&iRc, pTempMessage);
		if (pFileNameA == NULL)
		{
			if (pMyContext->strFtpFileName == NULL || strlen(pMyContext->strFtpFileName) == 0)
			{
				iFtpRc = 1;
				cciLog(&iRc, CCI_LOG_ERROR, __FILE__, __LINE__, "Evaluate",
					L"MQSIV2_FTPSEND", MSG_NO_FILENAME,
					"No file name supplied in tree or as attribute",
					(char *)0);
			}
			else
			{
				iFtpRc = FtpSend(pMyContext, pMyContext->strFtpFileName, pData, iSize);
			}
		}
		else
		{
			iFtpRc = FtpSend(pMyContext, pFileNameA, pData, iSize);
			free(pFileNameA);
		}
	}
	cniDeleteMessage(&iRc, pTempMessage);

	// If the FTP was successful, propogate the message through the "out" terminal
	// else through the "failure" terminal.
	if (iFtpRc == 0)
	{
		cniPropagate(&iRc, pMyContext->outTerminal, 
			inDestinationList, inExceptionList, inMessage);
	}
	else
	{
		cniPropagate(&iRc, pMyContext->failureTerminal, 
			inDestinationList, inExceptionList, inMessage);
	}

	if (iRc != 0)
	{
		mqsixLog("An exception was reported on return from propogate through 'out': %d\n", iRc);
		cciRethrowLastException(&iRc);
		return;
	}
	mqsixLog("<< Evaluate\n");
	return;
}

///////////////////////////////////////////////////////////
// bipGetMessageflowNodeFactory
// 
// Entry point into the FtpSend node.  Called by the broker when
// the node is initially loaded.


void __declspec(dllexport) *bipGetMessageflowNodeFactory()
{
	static CNI_VFT	virtualFunctionTable;
	CciChar			factoryName[] = L"FtpSendFactory";
	int				iRc;
	void			*factoryObject;

#ifdef _WIN32
	mqsixRegisterMessages(g_module, "MQSIV2_FTPSEND");
#endif
	mqsixLog("\n");
	mqsixLog("************************\n");
	mqsixLog("*** Loading Node LIL ***\n");
	mqsixLog("************************\n");
	mqsixLog(">> Creating nodeFactory: %S, lil build: %s, CCI Version: %d.%d\n",
		factoryName, __TIMESTAMP__, CCI_VERSION >> 16, CCI_VERSION & 0xFFFF);
	memset(&virtualFunctionTable, 0, sizeof(virtualFunctionTable));
	virtualFunctionTable.iFpCreateNodeContext	= CreateNodeContext;
	virtualFunctionTable.iFpDeleteNodeContext	= DeleteNodeContext;
	virtualFunctionTable.iFpGetAttributeName	= GetAttributeName;
	virtualFunctionTable.iFpSetAttribute		= SetAttribute;
	virtualFunctionTable.iFpGetAttribute		= GetAttribute;
	virtualFunctionTable.iFpEvaluate			= Evaluate;

	factoryObject = cniCreateNodeFactory(&iRc, factoryName);
	cniDefineNodeClass(&iRc, factoryObject, L"FtpSendNode", &virtualFunctionTable); 

	return(factoryObject);
}
